import specialTypeObj from './SpecialType'

function create_context_function_template (eval_string, context) {
  return `
  return function (context) {
    "use strict";
    ${Object.keys(context).length > 0
      ? `let ${Object.keys(context).map((key) => ` ${key} = context['${key}']`)};`
      : ``
    }
    return ${eval_string};
  }                                                                                                                   
  `
}

function make_context_evaluator(eval_string, context) {
  let template = create_context_function_template(eval_string, context)
  let functor = Function(template)
  return functor()
}

function evalInScope (js, context) {
  const evaluator = make_context_evaluator(js, context)
  return evaluator(context)
}

function evalScript (expression) {
  try {
    return evalInScope(expression, {
      wx: $window.wx
    })
  } catch (error) {
    return unDefinedError(expression)
  }
}

function commonType (res, runtime) {
  const className = typeof res
  if (res === null) {
    return {
      type: 'object',
      value: 'null',
      subtype: 'null'
    }
  }
  const special = specialTypeObj.getSpecialType(res)
  if (special) {
    if (special === 'Map' || special === 'Set') {
      return specialTypeObj.mapAndSetResult(special, res, runtime)
    }
    if (special === 'ArrayBuffer') {
      return specialTypeObj.arrayBufferResult(res)
    }
  }
  if (className === 'number') {
    return {
      type: 'number',
      value: res,
      description: res + ''
    }
  } else if (className === 'string') {
    return {
      type: 'string',
      value: res
    }
  } else if (className === undefined) {
    return {
      type: 'undefined'
    }
  } else if (className === 'boolean') {
    return {
      type: 'boolean',
      value: res
    }
  } else if (className === 'function') {
    return {
      type: "function",
      value: ""
    }
  } else if (className === 'object') {
    return commonTypeObj(res, runtime)
  }
}

function EvaluateRes (res, properties = [], key, runtime, tag) {
  const className = typeof res
  let result = commonType(res, runtime)
  if (className === 'object') {
    if (key && !tag) {
      if (res === null) {
        result = { type: 'object', value: 'null', subtype: 'null' }
      } else {
        result = { type: 'object', value: 'Object' }
      }
    } else {
      let __objectId__ = null
      if (res && res.__objectId__) {
        __objectId__ = res.__objectId__
      }
      const isSpecialType = specialTypeObj.getSpecialType(res)
      if (isSpecialType) { // 特殊对象如：map set 之类的处理
        return result
      }
      result = {
        type: 'object',
        className: 'Object',
        description: 'Object',
        objectId: runtime.setItemObjectId(__objectId__, res),
        preview: {
          type: 'object',
          description: 'Object',
          overflow: false,
          properties: properties
        }
      }
      for (const key in res) {
        if (Object.hasOwnProperty.call(res, key)) {
          const element = res[key]
          if (key === '__objectId__') {
            break
          }
          let tmp = null
          if (!tag) {
            tmp = EvaluateRes(element, properties, key, runtime, tag)
            properties.push(tmp)
          } else {
            if (typeof element === 'object') {
              tmp = EvaluateRes(element, properties, key, runtime, tag)
              properties.push(tmp)
            }
          }
        }
      }
    }
  } else if (className === 'function') {
    result.name = res.name
  } else {

  }
  if (key) {
    result.name = key
  }
  return result
}

function EvaluatePropertiesRes (data, runtime) {
  let result = []
  const isSpecialType = specialTypeObj.getSpecialType(data)
  const unsupported = specialTypeObj.unsupportedObjects(data)
  if (unsupported) {
    result = []
  } else {
    if (isSpecialType && data.forEach) {
      let index = 0
      data.forEach((ele, key) => {
        if (key !== '__objectId__') {
          if (isSpecialType === 'Set') {
            index++
            setEvaluatePropertiesResult(ele, result, index, runtime)
          } else {
            setEvaluatePropertiesResult(ele, result, key, runtime)
          }
        }
      })
    } else {
      for (const key in data) {
        if (Object.hasOwnProperty.call(data, key)) {
          if (key !== '__objectId__') {
            const element = data[key]
            setEvaluatePropertiesResult(element, result, key, runtime)
          }
        }
      }
    }
  }

  const internalProperties =  [{
		"name": "[[Prototype]]",
		"value": {
			"type": "object",
			"className": "Object",
			"description": "Object",
			"objectId": runtime.setItemObjectId(data.__objectId__, data)
		}
	}]
  return {
    result: result,
    internalProperties: internalProperties
  }
}

function setEvaluatePropertiesResult (element, result, key, runtime) {
  let tmp = commonType(element, runtime)
  result.push({
    "writable": true,
    "configurable": true,
    "enumerable": true,
    "isOwn": true,
    name: key + '' || '',
    value: tmp
  })
}

function commonTypeObj (element, runtime) {
  let tmp = null
  let __objectId__ = null
  let name = null
  const value = 'Object'
  if (element === null) {
    tmp = {
      type: 'object',
      subtype: 'null',
      value: null
    }
  } else {
    __objectId__ = element.__objectId__
    name = Object.keys(element)[0]
    tmp = {
      type: "object",
      className: "Object",
      description: "Object",
      objectId: runtime.setItemObjectId(__objectId__, element),
      preview: {
        type: "object",
        description: "Object",
        overflow: false,
        properties: [
          {
            name: name,
            type: "object",
            value: value
          }
        ]
      }
    }
  }
  return tmp
}

function unDefinedError (expression) {
  return {
    exceptionDetails: {
      "exceptionId": 578,
      "text": "Uncaught",
      "lineNumber": 0,
      "columnNumber": 0,
      "scriptId": "1713",
      "stackTrace": {
        "callFrames": [
          {
            "functionName": "",
            "scriptId": "1713",
            "url": "",
            "lineNumber": 0,
            "columnNumber": 0
          }
        ]
      },
      "exception": {
        "type": "object",
        "subtype": "error",
        "className": "ReferenceError",
        "description": `ReferenceError: ${expression} is not defined\n    at <anonymous>:1:1`,
        "objectId": "-5318114322147940757.8.20360"
      }

    },
    result: {
      "type": "object",
      "subtype": "error",
      "className": "ReferenceError",
      "description": `ReferenceError: ${expression} is not defined\n    at <anonymous>:1:1`,
      "objectId": "-5318114322147940757.8.20359",
      "preview": {
        "type": "object",
        "subtype": "error",
        "description": `ReferenceError: ${expression} is not defined\n    at <anonymous>:1:1`,
        "overflow": false,
        "properties": [
          {
            "name": "stack",
            "type": "string",
            "value": `ReferenceError: ${expression} is not defined\n    at <anonymous>:1:1`
          },
          {
            "name": "message",
            "type": "string",
            "value": `${expression} is not defined`
          }
        ]
      }
    }
  }
}

function getArgByArguments (arg, runtime) {
  const res = []
  for (const key in arg) {
    if (Object.hasOwnProperty.call(arg, key)) {
      let result = {}
      const element = arg[key]
      const expr = createExprByArg(element)
      const obj = evalScript(expr)
      if (typeof element === 'object') {
        result = specialType(runtime, element, element)
      } else {
        result = specialType(runtime, obj, element)
      }
      res.push(result)
    }
  }
  return res
}

/**
 * 特殊类型的处理(map、set)
 * @param {*} runtime 
 * @param {*} obj 
 * @param {*} element 
 * @returns 
 */
function specialType (runtime, obj, element) {
  const isSpecialType = specialTypeObj.getSpecialType(element)
  if (isSpecialType) {
    setObjItemId(runtime, element)
    return EvaluateRes(element, [], null, runtime)
  }
  if (typeof obj === 'object') {
    setObjItemId(runtime, obj)
    return EvaluateRes(obj, [], null, runtime)
  }
  return EvaluateRes(obj, [], null, runtime)
}

// 通过Arguments生成控制台需要的表达式(简单类型的处理)
function createExprByArg (element) {
  let expr = element
  const t = typeof element
  if (t === 'string') {
    expr = `"${element}"`
  } else if (t === 'object') {
    expr = `(${JSON.stringify(element)})`
  }
  return expr
}

function setObjItemId (runtime, obj) {
  const itemId = runtime.setItem(obj)
  obj.__objectId__ = `${itemId}_0`
  // runtime.setObjectId(obj, itemId)
}

const isIOS = $window.navigator.userAgent.indexOf('iPhone') !== -1 || $window.navigator.userAgent.indexOf('iPad') !== -1

export default {
  EvaluateRes,
  evalScript,
  EvaluatePropertiesRes,
  getArgByArguments,
  isIOS
}
